import html
import pathlib

from fastapi import APIRouter, Depends, HTTPException
from starlette import status

from api import db, get_conn
from model.bcd_request import BCDRequestModel, BCDProtectionRequirements, BCDSystemType, IPAddressVersion
from model.settings import settings
from model.wapi.cntl import APIToken, Mgr
from util.auth import check_auth
from util.util import render_jinja_template, send_email
from util.wapi_util import execute_wapi_function

router = APIRouter(
    prefix="/api/bcd_request",
    tags=["bcd_request"]
)
js_mods = ['bcdrequest']


@router.post('')
async def handle_request(bcd_request: BCDRequestModel, token: APIToken = Depends(check_auth), conn=Depends(get_conn)):
    user = Mgr.check_token(conn, token)

    host_is_prod = db.host_omdl.OP_ENV_IS_PROD
    if not host_is_prod and not user.has_permission(conn, 'wapi.devel_versions_user'):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='This is the development site! '
                   'Please use patch request on the production site (https://netvs.scc.kit.edu)'
        )

    ou_request = execute_wapi_function(conn, [
        {"idx": "own_mgr2ou_list", "name": "cntl.mgr2ou.list", "old": {"is_own": True}},
        {"idx": "unit_list", "name": "org.unit.list", "inner_join_ref": {"own_mgr2ou_list": "default"},
         "old": {"sorting_params_list": ["sub_position"]}},
        {"idx": "unit_list_superset", "name": "org.unit.list",
         "inner_join_ref": {"unit_list": "api_func_org_unit_hierarchy_gfk_list_is_superset"}},
        {"name": "cntl.mgr2ou.list", "inner_join_ref": {"unit_list_superset": "default"}, "idx": "mgr2ou_list"},
        {"name": "cntl.mgr.list", "inner_join_ref": {"mgr2ou_list": "default"}, "idx": "mgr_list"},
        {"name": "dns.fqdn2ou.list", "inner_join_ref": {"unit_list_superset": "default"}, "idx": "fqdn2ou_list"},
        {"name": "nd.bcd2ou.list", "inner_join_ref": {"unit_list_superset": "default"}, "idx": "bcd2ou_list"},
        {"name": "nd.bcd.list", "inner_join_ref": {"bcd2ou_list": "default"}, "idx": "bcd_list"},
        {"name": "org.unit_type.list", "inner_join_ref": {"unit_list": "default"}, "idx": "unit_type_list"}],
                                       user=user.login_name)

    ous = [item["short_name"] for item in ou_request["unit_list"]]

    if bcd_request.oe not in ous:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='User is not member of the specified organizational unit.'
        )

    group_request = execute_wapi_function(conn, [
        {"idx": "current_ou", "name": "org.unit.list", "old": {"short_name": bcd_request.oe}},
        {"idx": "child_ous", "name": "org.unit.list",
         "inner_join_ref": {"current_ou": "api_func_org_unit_hierarchy_gfk_list_is_superset"}},
        {"idx": "parent_ous", "name": "org.unit.list", "old": {"sorting_params_list": ["tree_level DESC"]},
         "inner_join_ref": {"current_ou": "api_func_org_unit_hierarchy_gfk_list_is_subset"}},
        {"name": "cntl.mgr2ou.list", "inner_join_ref": {"current_ou": "default"}, "idx": "mgr2ou_list"},
        {"name": "cntl.mgr.list", "inner_join_ref": {"mgr2ou_list": "default"}, "idx": "mgr_list"},
        {"name": "dns.fqdn2ou.list", "inner_join_ref": {"current_ou": "default"}, "idx": "fqdn2ou_list"},
        {"name": "nd.bcd2ou.list", "inner_join_ref": {"current_ou": "default"}, "idx": "bcd2ou_list"},
        {"name": "nd.bcd.list", "inner_join_ref": {"bcd2ou_list": "default"}, "idx": "bcd_list"},
        {"name": "cntl.group.list", "inner_join_ref": {"child_ous": "default"},
         "old": {"sorting_params_list": ["name"]}, "idx": "group_list"},
        {"idx": "recursive_fqdn2ou_list", "name": "dns.fqdn2ou.list", "inner_join_ref": {"parent_ous": "default"},
         "old": {"is_recursive": True}},
        {"idx": "recursive_bcd2ou_list", "name": "nd.bcd2ou.list", "inner_join_ref": {"parent_ous": "default"},
         "old": {"is_recursive": True}},
        {"idx": "recursive_mgr2ou_list", "name": "cntl.mgr2ou.list", "inner_join_ref": {"parent_ous": "default"}},
        {"idx": "recursive_mgr_list", "name": "cntl.mgr.list", "inner_join_ref": {"recursive_mgr2ou_list": "default"}},
        {"idx": "recursive_bcd_list", "name": "nd.bcd.list", "inner_join_ref": {"recursive_bcd2ou_list": "default"}},
        {"name": "org.unit_type.list", "inner_join_ref": {"current_ou": "default"}, "idx": "unit_type_list"}],
                                          user=user.login_name)

    groups = [item["name"] for item in group_request["group_list"]]

    if bcd_request.group not in groups:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Unknown group specified.'
        )

    if validate_bcd_request(bcd_request):
        mailstatus = send_bcd_request(bcd_request=bcd_request, mgr=user)
        if mailstatus:
            send_bcd_request_confirmation(bcd_request=bcd_request, mgr=user, receiver=user.email)
            return {'result': 'success'}

    raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)


def validate_bcd_request(bcd_request: BCDRequestModel) -> bool:
    if (bcd_request.protection_requirement == BCDProtectionRequirements.SPECIAL and
            bcd_request.protection_requirement_note is None):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='A description of the special protection requirement is required.'
        )
    elif (bcd_request.protection_requirement == BCDProtectionRequirements.SPECIAL and
          bcd_request.protection_requirement_note is not None and
          len(bcd_request.protection_requirement_note.strip()) == 0):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='A description of the special protection requirement is required.'
        )

    if bcd_request.type_of_system == BCDSystemType.OTHER and bcd_request.type_of_system_note is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='A detailed description of the system type other must be provided.'
        )
    elif (bcd_request.type_of_system == BCDSystemType.OTHER and bcd_request.type_of_system_note is not None and
          len(bcd_request.type_of_system_note.strip()) == 0):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='A detailed description of the system type other must be provided.'
        )

    if bcd_request.type_of_system == BCDSystemType.SERVER and bcd_request.load_balancer is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='You must provide load balancer info for server systems.'
        )

    if (bcd_request.type_of_system == BCDSystemType.SERVER and
            bcd_request.protection_requirement == BCDProtectionRequirements.LEVEL_2 and
            bcd_request.load_balancer is True):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Load balancers are not supported with level two protection requirements.'
        )

    if len(bcd_request.bcd_name.strip()) == 0:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='BCD cannot be empty.'
        )

    if (bcd_request.type_of_system != BCDSystemType.SERVER and bcd_request.access_from_internet and
            bcd_request.access_from_internet_note is None):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='A detailed description must be provided for access from the internet onto non server networks.'
        )
    elif (bcd_request.type_of_system != BCDSystemType.SERVER and bcd_request.access_from_internet and
          bcd_request.access_from_internet_note is not None and
          len(bcd_request.access_from_internet_note.strip()) == 0):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='A detailed description must be provided for access from the internet onto non server networks.'
        )

    if bcd_request.type_of_subnet == IPAddressVersion.DUAL_STACK and bcd_request.ipv4_subnet_size < 0:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Invalid subnet size.'
        )
    elif bcd_request.type_of_subnet == IPAddressVersion.DUAL_STACK and bcd_request.ipv4_subnet_size > 32:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Invalid subnet size.'
        )
    elif bcd_request.type_of_subnet == IPAddressVersion.DUAL_STACK and 29 < bcd_request.ipv4_subnet_size <= 32:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='No subnets smaller than /29 can be provided.'
        )
    elif bcd_request.type_of_subnet == IPAddressVersion.DUAL_STACK and bcd_request.ipv4_subnet_size < 24:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='No subnets larger than /24 can be provided.'
        )
    elif (bcd_request.type_of_subnet == IPAddressVersion.DUAL_STACK and bcd_request.ipv4_subnet_size == 24 and
          bcd_request.ipv4_subnet_size_explanation is None):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='A detailed description of the demand for a /24 subnet must be provided.'
        )
    elif (bcd_request.type_of_subnet == IPAddressVersion.DUAL_STACK and bcd_request.ipv4_subnet_size == 24 and
          bcd_request.ipv4_subnet_size_explanation is not None and
          len(bcd_request.ipv4_subnet_size_explanation.strip()) == 0):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='A detailed description of the demand for a /24 subnet must be provided.'
        )

    if bcd_request.dhcp_pool and not bcd_request.dhcp:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='DHCP must be enabled to support the DHCP pool.'
        )

    if bcd_request.dhcp and bcd_request.dhcp_pool and bcd_request.dhcp_pool_size < 2:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='DHCP pools smaller than 2 devices are not offered.'
        )
    elif (bcd_request.dhcp and bcd_request.dhcp_pool and
          bcd_request.dhcp_pool_size > 2 ** (32 - bcd_request.ipv4_subnet_size) - 5):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='This subnet does not support this size of a DHCP pool.'
        )

    if bcd_request.dhcp and bcd_request.own_dhcp_server:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='The provided DHCP server cannot be used concurrently to your own DHCP server.'
        )

    if bcd_request.type_of_system == BCDSystemType.VPN_WIFI_ONLY and bcd_request.location_vpn_wifi_to_vlan is False:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='The VPN2vlan/Wifi2vlan client location must be selected for VPN2vlan/Wifi2vlan only system types.'
        )

    if bcd_request.location_vpn_wifi_to_vlan and bcd_request.vpn2vlan is False and bcd_request.wifi2vlan is False:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='At least VPN2vlan or Wifi2vlan must be enabled for the VPN2vlan/Wifi2vlan client location.'
        )
    elif bcd_request.wifi2vlan and bcd_request.location_vpn_wifi_to_vlan is False:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='The VPN2vlan/Wifi2vlan client location must be selected for Wifi2vlan.'
        )
    elif bcd_request.vpn2vlan and bcd_request.location_vpn_wifi_to_vlan is False:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='The VPN2vlan/Wifi2vlan client location must be selected for VPN2vlan.'
        )

    return True


def send_bcd_request(bcd_request: BCDRequestModel, mgr: Mgr, receiver=settings.bcd_request_email_receivers) -> bool:
    try:
        host_mode = db.host_omdl.OP_ENV_MODE.upper()

        path = pathlib.Path(__file__).parent.parent.resolve()
        TEMPLATE = 'templates/bcd_request_template.j2'

        body = render_jinja_template(path, TEMPLATE,
                                     bcd_request=bcd_request,
                                     mgr=mgr,
                                     host_mode=host_mode,
                                     )

        subject = (f'BCD Request for {html.escape(bcd_request.oe)} by '
                   f'{html.escape(mgr.first_name)} {html.escape(mgr.last_name)}')
        if host_mode:
            subject = f'[{host_mode}] {subject}'

        # hacky way to intercept mails for development. In production, the assignment should be used
        if settings.bcd_request_override_mail_receiver is not None:
            receiver = settings.bcd_request_override_mail_receiver

        send_email(to=receiver,
                   sender=settings.patch_request_email_sender,
                   reply_to=mgr.email,
                   subject=subject,
                   body=body,
                   )
        return True
    except Exception as e:
        raise e


def send_bcd_request_confirmation(bcd_request: BCDRequestModel, mgr: Mgr, receiver) -> bool:
    try:
        host_mode = db.host_omdl.OP_ENV_MODE.upper()

        path = pathlib.Path(__file__).parent.parent.resolve()
        TEMPLATE = 'templates/bcd_request_template.j2'

        body = render_jinja_template(path, TEMPLATE,
                                     bcd_request=bcd_request,
                                     mgr=mgr,
                                     host_mode=host_mode,
                                     confirmation=True
                                     )

        subject = f'Confirmation for BCD Request for {html.escape(bcd_request.oe)}'
        if host_mode:
            subject = f'[{host_mode}] {subject}'

        # hacky way to intercept mails for development. In production, the assignment should be used
        if settings.bcd_request_override_mail_receiver is not None:
            receiver = settings.bcd_request_override_mail_receiver

        send_email(to=receiver,
                   sender=settings.bcd_request_email_sender,
                   reply_to=settings.bcd_request_email_sender,
                   subject=subject,
                   body=body
                   )
        return True
    except Exception as e:
        raise e
