import datetime
import importlib
import sys
import traceback

import uvicorn
from fastapi import FastAPI, Request
from starlette import status
from starlette.responses import JSONResponse

import api.misc
from model.settings import settings

app = FastAPI()
app.openapi_version = '3.0.2'


def add_api_module(module_name: str):
    module = importlib.import_module('api.' + module_name)
    app.include_router(module.router)


app.include_router(api.misc.router)
for mod in settings.loaded_modules:
    add_api_module(mod)


# TODO: rename method
@app.middleware('http')
async def add_database_to_request(request: Request, call_next):
    request.state.success_callbacks = []
    request.state.error_callbacks = []
    request.state.common_callbacks = []

    # noinspection PyBroadException
    try:
        response = await call_next(request)
        for callback in request.state.success_callbacks:
            callback()
        csp = "default-src 'self'; script-src 'self'"
        response.headers['Content-Security-Policy'] = csp
        response.headers['X-Content-Security-Policy'] = csp
        response.headers['X-WebKit-CSP'] = csp
        response.headers['X-Frame-Options'] = "SAMEORIGIN"
        response.headers['X-Xss-Protection'] = "1; mode=block"
        response.headers['X-Content-Type-Options'] = "nosniff"
        response.headers['Referrer-Policy'] = "same-origin"

        return response
    except Exception as e:
        for callback in request.state.error_callbacks:
            callback()
        print('Ohh no. An exception occurred. Sad unicorn noises :(\n', e, file=sys.stderr)
        traceback.print_exc()
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            content={
                'message': 'Oops! Something went wrong. There goes a rainbow...',
                'reason': str(e),
                'path': request.url.path,
                'query': request.url.query,
                'time': datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            },
        )
    finally:
        for callback in request.state.common_callbacks:
            callback()


# Mockup
@app.get('/api/alerts.json')
async def alerts_json():
    return [
        {
            'title': {'en': 'Development server', 'de': 'Entwicklungsumgebung'},
            'content': {
                'en': 'This alert comes from the middleware. It should be served by a dedicated file!',
                'de': 'Dieser Text wird von der Middleware bereitgestellt, es sollte aber eine eigene Datei sein!',
            }
        }
    ]


if __name__ == '__main__':
    print(
        """
        ██╗    ██╗ █████╗ ██████╗ ███╗   ██╗██╗███╗   ██╗ ██████╗
        ██║    ██║██╔══██╗██╔══██╗████╗  ██║██║████╗  ██║██╔════╝
        ██║ █╗ ██║███████║██████╔╝██╔██╗ ██║██║██╔██╗ ██║██║  ███╗
        ██║███╗██║██╔══██║██╔══██╗██║╚██╗██║██║██║╚██╗██║██║   ██║
        ╚███╔███╔╝██║  ██║██║  ██║██║ ╚████║██║██║ ╚████║╚██████╔╝
         ╚══╝╚══╝ ╚═╝  ╚═╝╚═╝  ╚═╝╚═╝  ╚═══╝╚═╝╚═╝  ╚═══╝ ╚═════╝
          This is just a local development-server!
          Use gunicorn/uvicorn in production!
          !!! DO NOT USE THIS MAIN-METHOD IN ANY PRODUCTION INSTANCE !!!
        """
    )
    # Startup a crude and unsafe development server
    uvicorn.run(app, host='localhost', port=8000)
